use std::collections::HashMap;
use std::net::{IpAddr, Ipv4Addr};

use anyhow::{anyhow, Result};
use byteorder::{BigEndian, ByteOrder};
use log::*;
use trust_dns_proto::op::{
    header::MessageType, op_code::OpCode, response_code::ResponseCode, Message,
};
use trust_dns_proto::rr::{
    dns_class::DNSClass, record_data::RData, record_type::RecordType, resource::Record,
};

pub struct FakeDns {
    map: HashMap<u32, String>,
    cursor: u32,
    min_cursor: u32,
    max_cursor: u32,
    ttl: u32,
    exclude_domains: Vec<String>,
}

impl FakeDns {
    pub fn new() -> Self {
        let min_cursor = Self::ip_to_u32(&Ipv4Addr::new(173, 255, 0, 0));
        let max_cursor = Self::ip_to_u32(&Ipv4Addr::new(173, 255, 4, 255));

        FakeDns {
            map: HashMap::new(),
            cursor: min_cursor,
            min_cursor,
            max_cursor,
            ttl: 1,
            exclude_domains: Vec::new(),
        }
    }

    pub fn exclude(&mut self, domain: String) {
        self.exclude_domains.push(domain);
    }

    fn allocate_ip(&mut self, domain: &String) -> Ipv4Addr {
        self.map.insert(self.cursor, domain.to_owned());
        let ip = Self::u32_to_ip(self.cursor);
        self.cursor += 1;
        if self.cursor > self.max_cursor {
            self.cursor = self.min_cursor;
        }
        ip
    }

    pub fn query_domain(&mut self, ip: &IpAddr) -> Option<String> {
        let ip = match ip {
            IpAddr::V4(ip) => ip,
            _ => return None,
        };
        match self.map.get(&Self::ip_to_u32(ip)) {
            Some(v) => Some(v.clone()),
            None => None,
        }
    }

    pub fn generate_fake_response(&mut self, request: &[u8]) -> Result<Vec<u8>> {
        let req = Message::from_vec(request)?;

        if req.queries().len() == 0 {
            return Err(anyhow!("no queries in this DNS request"));
        }

        let query = &req.queries()[0];
        if query.query_class() != DNSClass::IN {
            return Err(anyhow!("unsupported query class {}", query.query_class()));
        }

        let t = query.query_type();
        if t != RecordType::A && t != RecordType::AAAA && t != RecordType::Unknown(65) {
            return Err(anyhow!(
                "unsupported query record type {:?}",
                query.query_type()
            ));
        }

        let raw_name = query.name();

        // TODO check if a valid domain
        let domain = if raw_name.is_fqdn() {
            let fqdn = raw_name.to_ascii();
            fqdn[..fqdn.len() - 1].to_string()
        } else {
            raw_name.to_ascii()
        };

        for ex in &self.exclude_domains {
            if domain.contains(ex) {
                return Err(anyhow!("domain {} excluded", domain));
            }
        }

        let ip = self.allocate_ip(&domain);
        debug!("allocate {} for {}", &ip, &domain);

        let mut resp = Message::new();

        // sets the response according to request
        // https://github.com/miekg/dns/blob/f515aa579d28efa1af67d9a62cc57f2dfe59da76/defaults.go#L15
        resp.set_id(req.id())
            .set_message_type(MessageType::Response)
            .set_op_code(req.op_code());

        if resp.op_code() == OpCode::Query {
            resp.set_recursion_desired(req.recursion_desired())
                .set_checking_disabled(req.checking_disabled());
        }
        resp.set_response_code(ResponseCode::NoError);
        if req.queries().len() > 0 {
            resp.add_query(query.clone());
        }

        if query.query_type() == RecordType::A {
            let mut ans = Record::new();
            ans.set_name(raw_name.clone())
                .set_rr_type(RecordType::A)
                .set_ttl(self.ttl)
                .set_dns_class(DNSClass::IN)
                .set_rdata(RData::A(ip));
            resp.add_answer(ans);
        }

        Ok(resp.to_vec()?)
    }

    pub fn is_fake_ip(&self, ip: &IpAddr) -> bool {
        let ip = match ip {
            IpAddr::V4(ip) => ip,
            _ => return false,
        };
        let ip = Self::ip_to_u32(ip);
        if ip >= self.min_cursor && ip <= self.max_cursor {
            true
        } else {
            false
        }
    }

    fn u32_to_ip(ip: u32) -> Ipv4Addr {
        Ipv4Addr::from(ip)
    }

    fn ip_to_u32(ip: &Ipv4Addr) -> u32 {
        BigEndian::read_u32(&ip.octets())
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use std::net::{IpAddr, Ipv4Addr};

    #[test]
    fn test_u32_to_ip() {
        let ip1 = Ipv4Addr::new(127, 0, 0, 1);
        let ip2 = FakeDns::u32_to_ip(2130706433u32);
        assert_eq!(ip1, ip2);
    }

    #[test]
    fn test_ip_to_u32() {
        let ip = Ipv4Addr::new(127, 0, 0, 1);
        let ip1 = FakeDns::ip_to_u32(&ip);
        let ip2 = 2130706433u32;
        assert_eq!(ip1, ip2);
    }
}
